@@ -9,17 +9,19 @@ share them or access application attributes.
 """
 
 import datetime
+import hashlib
 import math
 import os
 import sys
 import time
 import xml.etree.ElementTree as ET
+from concurrent.futures import Future, ThreadPoolExecutor
 
 import requests
 import six
 from six.moves.urllib import parse
 
-from owncloud import utils
+from . import utils
 
 
 class ResponseError(Exception):
@@ -368,9 +370,9 @@ class Client(object):
             url_components = parse.urlparse(self.url)
             if self._dav_endpoint_version == 1:
                 self._davpath = url_components.path + \
-                    'remote.php/dav/files/' + parse.quote(user_id)
+                                'remote.php/dav/files/' + parse.quote(user_id)
                 self._webdav_url = self.url + \
-                    'remote.php/dav/files/' + parse.quote(user_id)
+                                   'remote.php/dav/files/' + parse.quote(user_id)
             else:
                 self._davpath = url_components.path + 'remote.php/webdav'
                 self._webdav_url = self.url + 'remote.php/webdav'
@@ -413,6 +415,20 @@ class Client(object):
         destination = '/' + os.path.basename(file_name)
         return self.put_file(destination, file_name)
 
+    def exists(self, path):
+        """
+        returns true if the file exists, otherwise false
+        """
+        res = self._session.request(
+            'HEAD',
+            self._webdav_url + parse.quote(self._encode_string(path))
+        )
+        if res.status_code < 300:
+            return True
+        if res.status_code == 404:
+            return False
+        raise HTTPResponseError(res)
+
     def file_info(self, path, properties=None):
         """Returns the file info for the given remote file
 
@@ -436,12 +452,12 @@ class Client(object):
             data = ET.tostring(root)
         else:
             data = None
-        res = self._make_dav_request('PROPFIND', path, headers={'Depth': '0'}, data=data)
+        res = self._make_dav_request('PROPFIND', path, headers={
+                                     'Depth': '0'}, data=data)
         if res:
             return res[0]
         return None
 
-
     def list(self, path, depth=1, properties=None):
         """Returns the listing/contents of the given remote directory
 
@@ -473,7 +489,8 @@ class Client(object):
         else:
             data = None
 
-        res = self._make_dav_request('PROPFIND', path, headers=headers, data=data)
+        res = self._make_dav_request(
+            'PROPFIND', path, headers=headers, data=data)
         # first one is always the root, remove it from listing
         if res:
             return res[1:]
@@ -536,7 +553,7 @@ class Client(object):
         """
         remote_path = self._normalize_path(remote_path)
         url = self.url + 'index.php/apps/files/ajax/download.php?dir=' \
-            + parse.quote(remote_path)
+              + parse.quote(remote_path)
         res = self._session.get(url, stream=True)
         if res.status_code == 200:
             if local_file is None:
@@ -626,20 +643,32 @@ class Client(object):
         if not local_directory.endswith('/'):
             local_directory += '/'
 
-        basedir = os.path.basename(local_directory[0: -1]) + '/'
+        basedir = os.path.basename(local_directory[:-1]) + '/'
         # gather files to upload
         for path, _, files in os.walk(local_directory):
             gathered_files.append(
                 (path, basedir + path[len(local_directory):], files)
             )
 
-        for path, remote_path, files in gathered_files:
-            self.mkdir(target_path + remote_path + '/')
-            for name in files:
-                if not self.put_file(target_path + remote_path + '/',
-                                     path + '/' + name, **kwargs):
+        futures = []
+        executor = ThreadPoolExecutor(max_workers=10, thread_name_prefix="owncloud_")
+        try:
+            for path, remote_path, files in gathered_files:
+                _path = os.path.join(target_path, remote_path)
+                if not kwargs.get('resume', False) or not self.exists(_path):
+                    self.mkdir(_path)
+                for name in files:
+                    res = self.put_file(_path, os.path.join(path, name), **kwargs, executor=executor)
+                    if isinstance(res, Future):
+                        futures.append(res)
+                    elif not res:
+                        return False
+            for future in futures:
+                if not future.result():
                     return False
-        return True
+            return True
+        finally:
+            executor.shutdown()
 
     def _put_file_chunked(self, remote_path, local_source_file, **kwargs):
         """Uploads a file using chunks. If the file is smaller than
@@ -654,7 +683,7 @@ class Client(object):
         """
         chunk_size = kwargs.get('chunk_size', 10 * 1024 * 1024)
         result = True
-        transfer_id = int(time.time())
+        now = int(time.time())
 
         remote_path = self._normalize_path(remote_path)
         if remote_path.endswith('/'):
@@ -668,25 +697,29 @@ class Client(object):
         file_handle.seek(0)
 
         headers = {}
+        st_mtime = int(stat_result.st_mtime)
         if kwargs.get('keep_mtime', True):
-            headers['X-OC-MTIME'] = str(int(stat_result.st_mtime))
-        global downloaded
-        downloaded = 0
+            headers['X-OC-MTIME'] = str(st_mtime)
+        # global downloaded
+        uploaded = 0
         console_size = 50
 
         def progress(chunk):
-            global downloaded
-            downloaded += chunk
-            done = int(console_size * downloaded / float(size))
-            percent = 100 * downloaded / float(size)
-            remain_time = (int(time.time()) - transfer_id) * \
-                ((100 - percent) / percent)
+            # global downloaded
+            nonlocal uploaded
+            uploaded += chunk
+            done = int(console_size * uploaded / float(size))
+            percent = 100 * uploaded / float(size)
+            remain_time = (int(time.time()) - now) * \
+                          ((100 - percent) / percent)
             if percent >= 100:
                 sys.stdout.write("\rUploading " + local_source_file + " [" + '#' * done + '.' * (
-                    console_size-done) + "] " + '{:.2f}'.format(percent) + "% Time: " + utils.format_time(int(time.time()) - transfer_id))
+                        console_size - done) + "] " + '{:.2f}'.format(percent) + "% Time: " + utils.format_time(
+                    int(time.time()) - now))
             else:
                 sys.stdout.write("\rUploading " + local_source_file + " [" + '#' * done + '.' * (
-                    console_size-done) + "] " + '{:.2f}'.format(percent) + "% ETA: " + utils.format_time(remain_time))
+                        console_size - done) + "] " + '{:.2f}'.format(percent) + "% ETA: " + utils.format_time(
+                    remain_time))
             sys.stdout.flush()
 
         if size == 0:
@@ -700,28 +733,73 @@ class Client(object):
 
         if chunk_count > 1:
             headers['OC-CHUNKED'] = '1'
+        index = 0
+        cache_file_name = hashlib.md5(
+            '{}-{}-{}-{}-{}'.format(
+                self._session.auth[0] if self._session else None,
+                local_source_file,
+                remote_path,
+                st_mtime,
+                chunk_count).encode('UTF-8')).hexdigest()
 
-        for chunk_index in range(0, int(chunk_count)):
+        if kwargs.get('resume', False):
+            if self.exists(remote_path):
+                prop = self.file_info(remote_path, ['oc:checksums'])
+                try:
+                    checksums = prop.attributes['{http://owncloud.org/ns}checksums']['{http://owncloud.org/ns}checksum'][0]
+                    remote_md5 = checksums.split(' ')[1].split(':')[1]
+                    md5 = hashlib.md5()
+                    with open(local_source_file, 'rb') as fr:
+                        while True:
+                            data = fr.read(1024*1024)
+                            if not data:
+                                break
+                            md5.update(data)
+                    local_md5 = md5.hexdigest()
+                    if remote_md5 == local_md5:
+                        if kwargs.get('show_progress_bar', True):
+                            progress(size)
+                        sys.stdout.write("\n")
+                        return True
+                except BaseException:
+                    pass
+            cached = utils.read_cache_file(cache_file_name, datetime.timedelta(hours=kwargs.get('cache_validity', 8)))
+            try:
+                index = int(cached)
+            except BaseException:
+                pass
+            file_handle.seek(index * chunk_size, os.SEEK_SET)
+            uploaded = index * chunk_size
+        for chunk_index in range(index, int(chunk_count)):
             data = file_handle.read(chunk_size)
             if chunk_count > 1:
                 chunk_name = '%s-chunking-%s-%i-%i' % \
-                             (remote_path, transfer_id, chunk_count,
+                             (remote_path, st_mtime, chunk_count,
                               chunk_index)
+                if chunk_index == chunk_count - 1 and kwargs.get('executor', None):
+                    def task():
+                        res = self._make_dav_request('PUT', chunk_name, data=data, headers=headers)
+                        if res:
+                            utils.delete_cache_file(cache_file_name)
+                            file_handle.close()
+                            if kwargs.get('show_progress_bar', True):
+                                progress(len(data) if chunk_count > 1 else size)
+                        sys.stdout.write("\n")
+                        return res
+                    return kwargs.get('executor').submit(task)
+                else:
+                    result = self._make_dav_request('PUT', chunk_name, data=data, headers=headers)
+                    utils.write_cache_file(cache_file_name, str(chunk_index + 1))
             else:
-                chunk_name = remote_path
-            if not self._make_dav_request(
-                    'PUT',
-                    chunk_name,
-                    data=data,
-                    headers=headers
-            ):
-                result = False
-                break
+                result = self._make_dav_request('PUT', remote_path, data=data, headers=headers)
+                file_handle.close()
             if kwargs.get('show_progress_bar', True):
                 progress(chunk_size if chunk_count > 1 else size)
-
-        file_handle.close()
+            if not result:
+                break
         sys.stdout.write("\n")
+        if result:
+            utils.delete_cache_file(cache_file_name)
         return result
 
     def mkdir(self, path):
@@ -1748,7 +1826,7 @@ class Client(object):
         :param s: str or unicode to encode
         :returns: encoded output as str
         """
-        if six.PY2 and isinstance(s, unicode): # noqa: F821
+        if six.PY2 and isinstance(s, unicode):  # noqa: F821
             return s.encode('utf-8')
         return s
 
@@ -1884,8 +1962,19 @@ class Client(object):
         file_attrs = {}
         attrs = dav_response.find('{DAV:}propstat')
         attrs = attrs.find('{DAV:}prop')
+
+        def recurse(element):
+            if element.text:
+                return element.text
+            result = {}
+            for child in element:
+                if not result.get(child.tag):
+                    result[child.tag] = []
+                result[child.tag].append(recurse(child))
+            return result
+
         for attr in attrs:
-            file_attrs[attr.tag] = attr.text
+            file_attrs[attr.tag] = recurse(attr)
 
         return FileInfo(href, file_type, file_attrs)
 
